from xdsl.irdl import *
from xdsl.ir import *


@dataclass
class Builtin:
    ctx: MLContext

    def __post_init__(self):
        self.ctx.register_attr(StringAttr)
        self.ctx.register_attr(FlatSymbolRefAttr)
        self.ctx.register_attr(SymbolNameAttr)
        self.ctx.register_attr(IntAttr)
        self.ctx.register_attr(IntegerAttr)
        self.ctx.register_attr(ArrayAttr)
        self.ctx.register_attr(VectorAttr)

        self.ctx.register_attr(FunctionType)
        self.ctx.register_attr(Float32Type)
        self.ctx.register_attr(IntegerType)
        self.ctx.register_attr(IndexType)

        self.ctx.register_op(ModuleOp)
        self.ctx.register_op(FuncOp)


@dataclass(frozen=True)
class StringAttr(Data):
    name = "string"
    data: str

    # parser should be of type parser.Parser
    @staticmethod
    def parse(parser) -> 'StringAttr':
        data = parser.parse_str_literal()
        return StringAttr(data)

    def print(self, printer) -> None:
        printer.print_string(f'"{self.data}"')

    @staticmethod
    def get(data: str) -> 'StringAttr':
        return StringAttr(data)


@irdl_attr_definition
class SymbolNameAttr(ParametrizedAttribute):
    name = "symbol_name"
    data = ParameterDef(StringAttr)

    @staticmethod
    def get(data: str) -> 'SymbolNameAttr':
        return SymbolNameAttr([StringAttr.get(data)])


@irdl_attr_definition
class FlatSymbolRefAttr(ParametrizedAttribute):
    name = "flat_symbol_ref"
    data = ParameterDef(StringAttr)

    @staticmethod
    def get(data: str) -> 'FlatSymbolRefAttr':
        return FlatSymbolRefAttr([StringAttr(data)])


@dataclass(frozen=True)
class IntAttr(Data):
    name = "int"
    data: int

    # parser should be of type parser.Parser
    @staticmethod
    def parse(parser: Any) -> 'IntAttr':
        data = parser.parse_int_literal()
        return IntAttr(data)

    # printer should be of type printer.Printer
    def print(self, printer) -> None:
        printer.print_string(f'{self.data}')

    @staticmethod
    def get(data: int) -> 'IntAttr':
        return IntAttr(data)


@irdl_attr_definition
class IntegerType:
    name = "integer_type"

    width = ParameterDef(IntAttr)

    @staticmethod
    def get(width: int) -> Attribute:
        return IntegerType([IntAttr.get(width)])


@irdl_attr_definition
class IndexType:
    name = "index"


@irdl_attr_definition
class IntegerAttr(ParametrizedAttribute):
    name = "integer"

    value = ParameterDef(IntAttr)
    typ = ParameterDef(AnyOf([IntegerType, IndexType]))

    @staticmethod
    def get(
        value: int, type: Attribute = IntegerType.get(64)) -> 'IntegerAttr':
        return IntegerAttr([IntAttr.get(value), type])


@dataclass(frozen=True)
class ArrayAttr(Data):
    name = "array"
    data: List[Attribute]

    @staticmethod
    def parse(parser) -> 'ArrayAttr':
        parser.parse_char("[")
        data = parser.parse_list(parser.parse_optional_attribute)
        parser.parse_char("]")
        return ArrayAttr.get(data)

    def print(self, printer) -> None:
        printer.print_string("[")
        printer.print_list(self.data, printer.print_attribute)
        printer.print_string("]")

    @staticmethod
    def get(data: List[Attribute]) -> 'ArrayAttr':
        return ArrayAttr(data)


# A constraint that enforces an ArrayData whose elements all satisfy
# the elem_constr
@dataclass
class ArrayOfConstraint(AttrConstraint):
    elem_constr: AttrConstraint

    def __init__(self, constr: Union[Attribute, typing.Type[Attribute],
                                     AttrConstraint]):
        self.elem_constr = attr_constr_coercion(constr)

    def verify(self, data: Data) -> None:
        if not isinstance(data, ArrayAttr):
            raise Exception(f"expected data ArrayData but got {data}")

        for e in data.data:
            self.elem_constr.verify(e)


@irdl_attr_definition
class VectorAttr(ParametrizedAttribute):
    name = "vector"
    data = ParameterDef(ArrayOfConstraint(IntegerAttr))

    @staticmethod
    def get(data: List[int]) -> 'VectorAttr':
        data_attr = [IntegerAttr.get(d) for d in data]
        return VectorAttr([ArrayAttr.get(data_attr)])


@irdl_attr_definition
class Float32Type:
    name = "f32"

    @staticmethod
    def get() -> Attribute:
        return Float32Type()


@irdl_attr_definition
class FunctionType:
    name = "fun"

    inputs = ParameterDef(ArrayOfConstraint(AnyAttr()))
    outputs = ParameterDef(ArrayOfConstraint(AnyAttr()))

    @staticmethod
    def get(inputs: List[Attribute], outputs: List[Attribute]) -> Attribute:
        return FunctionType([ArrayAttr.get(inputs), ArrayAttr.get(outputs)])


@irdl_op_definition
class FuncOp:
    name: str = "builtin.func"

    body = RegionDef()
    sym_name = AttributeDef(StringAttr)
    type = AttributeDef(Attribute)
    sym_visibility = AttributeDef(StringAttr)


@irdl_op_definition
class ModuleOp:
    name: str = "module"

    body = SingleBlockRegionDef()

    @property
    def ops(self) -> List[Operation]:
        return self.regions[0].blocks[0].ops



from __future__ import annotations
from xdsl.dialects.builtin import *
from typing import TypeVar

indentNumSpaces = 2


class Parser:
    def __init__(self, ctx: MLContext, _str: str):
        self._ctx: MLContext = ctx
        self._str: str = _str
        self._idx: int = 0
        self._ssaValues: Dict[str, SSAValue] = dict()
        self._blocks: Dict[str, Block] = dict()

    def skip_white_space(self) -> None:
        while self._idx < len(self._str):
            if self._str[self._idx].isspace():
                self._idx += 1
            # TODO: rewrite this hack to support comments
            elif self._idx < len(self._str) - 1 and self._str[
                    self._idx] == self._str[self._idx + 1] == '/':
                self.parse_while(lambda x: x != '\n', False)

            else:
                return

    def parse_while(self,
                    cond: Callable[[str], bool],
                    skip_white_space=True) -> str:
        if skip_white_space:
            self.skip_white_space()
        start_idx = self._idx
        while self._idx < len(self._str):
            char = self._str[self._idx]
            if not cond(char):
                return self._str[start_idx:self._idx]
            self._idx += 1
        return self._str[start_idx:]

    # TODO why two different functions, no nums in ident?
    def parse_optional_ident(self, skip_white_space=True) -> Optional[str]:
        res = self.parse_while(lambda x: x.isalpha() or x == "_" or x == ".",
                               skip_white_space=skip_white_space)
        if len(res) == 0:
            return None
        return res

    def parse_ident(self, skip_white_space=True) -> str:
        res = self.parse_optional_ident(skip_white_space=skip_white_space)
        if res is None:
            raise Exception("ident expected")
        return res

    def parse_alpha_num(self, skip_white_space=True) -> str:
        res = self.parse_while(lambda x: x.isalnum() or x == "_" or x == ".",
                               skip_white_space=skip_white_space)
        if len(res) == 0:
            raise Exception("alphanum expected")
        return res

    def parse_optional_str_literal(self) -> Optional[str]:
        parsed = self.parse_optional_char('"')
        if parsed is None:
            return None
        res = self.parse_while(lambda char: char != '"',
                               skip_white_space=False)
        self.parse_char('"')
        return res

    def parse_str_literal(self) -> str:
        res = self.parse_optional_str_literal()
        if res is None:
            raise Exception("string literal expected")
        return res

    def parse_optional_int_literal(self) -> Optional[int]:
        res = self.parse_while(lambda char: char.isnumeric())
        if len(res) == 0:
            return None
        return int(res)

    def parse_int_literal(self) -> int:
        res = self.parse_optional_int_literal()
        if res is None:
            raise Exception("int literal expected")
        return res

    def peek_char(self, char: str) -> Optional[bool]:
        self.skip_white_space()
        if self._idx == len(self._str):
            return None
        if self._str[self._idx] == char:
            return True
        return None

    def parse_optional_char(self, char: str) -> Optional[bool]:
        assert (len(char) == 1)
        res = self.peek_char(char)
        if res:
            self._idx += 1
        return res

    def parse_char(self, char: str) -> bool:
        assert (len(char) == 1)
        res = self.parse_optional_char(char)
        if res is None:
            raise Exception("'%s' expected" % char)
        return True

    def parse_string(self, contents: List[str]) -> bool:
        self.skip_white_space()
        for char in contents:
            if self._idx >= len(self._str):
                raise Exception("'%s' expected" % str)
            if self._str[self._idx] == char:
                self._idx += 1
        return True

    T = TypeVar('T')

    def parse_list(self,
                   parse_optional_one: Callable[[], Optional[T]],
                   delimiter=",") -> List[T]:
        assert (len(delimiter) <= 1)
        res = []
        one = parse_optional_one()
        if one is not None:
            res.append(one)
        while self.parse_optional_char(delimiter) if len(
                delimiter) == 1 else True:
            one = parse_optional_one()
            if one is None:
                return res
            res.append(one)
        return res

    def parse_optional_block_argument(
            self) -> Optional[Tuple[str, BlockArgument]]:
        name = self.parse_optional_ssa_name()
        if name is None:
            return None
        self.parse_char(":")
        typ = self.parse_attribute()
        # TODO how to get the id?
        return name, BlockArgument(typ, None, 0)

    def parse_optional_named_block(self) -> Optional[Block]:
        if self.parse_optional_char("^") is None:
            return None
        block_name = self.parse_alpha_num(skip_white_space=False)
        if block_name in self._blocks:
            block = self._blocks[block_name]
        else:
            block = Block()
            if self.parse_optional_char("("):
                tuple_list = self.parse_list(
                    self.parse_optional_block_argument)
                # TODO can we clean this up a bit?
                # Register the BlockArguments as ssa values and add them to
                # the block
                for (idx, res) in enumerate(tuple_list):
                    if res[0] in self._ssaValues:
                        raise Exception("SSA value %s is already defined" %
                                        res[0])
                    arg = res[1]
                    self._ssaValues[res[0]] = arg
                    arg.index = idx
                    arg.block = block
                    block.args.append(arg)

                self.parse_char(")")
            self._blocks[block_name] = block
        self.parse_char(":")
        for op in self.parse_list(self.parse_optional_op, delimiter=""):
            block.add_op(op)
        return block

    def parse_optional_region(self) -> Optional[Region]:
        if not self.parse_optional_char("{"):
            return None
        region = Region()

        if self.peek_char("^"):
            for block in self.parse_list(self.parse_optional_named_block,
                                         delimiter=""):
                region.add_block(block)
        else:
            region.add_block(Block())
            for op in self.parse_list(self.parse_optional_op, delimiter=""):
                region.blocks[0].add_op(op)
        self.parse_char("}")
        return region

    def parse_optional_ssa_name(self) -> Optional[str]:
        if self.parse_optional_char("%") is None:
            return None
        name = self.parse_alpha_num()
        return name

    def parse_optional_ssa_value(self) -> Optional[SSAValue]:
        name = self.parse_optional_ssa_name()
        if name is None:
            return None
        if name not in self._ssaValues:
            raise Exception("name '%s' does not refer to a SSA value" % name)
        return self._ssaValues[name]

    def parse_optional_result(self) -> Optional[Tuple[str, Attribute]]:
        name = self.parse_optional_ssa_name()
        if name is None:
            return None
        self.parse_char(":")
        typ = self.parse_attribute()
        return name, typ

    def parse_optional_results(self) -> Optional[List[Tuple[str, Attribute]]]:
        # One argument
        res = self.parse_optional_result()
        if res is not None:
            self.parse_char("=")
            return [res]

        # No arguments
        if self.parse_optional_char("(") is None:
            return None

        # Multiple arguments
        res = self.parse_list(lambda: self.parse_optional_result())
        self.parse_char(")")
        self.parse_char("=")
        return res

    def parse_optional_operand(self) -> Optional[SSAValue]:
        value = self.parse_optional_ssa_value()
        if value is None:
            return None
        self.parse_char(":")
        typ = self.parse_attribute()
        if value.typ != typ:
            raise Exception("type mismatch between %s and %s" %
                            (typ, value.typ))
        return value

    def parse_operands(self) -> List[Optional[SSAValue]]:
        self.parse_char("(")
        res = self.parse_list(lambda: self.parse_optional_operand())
        self.parse_char(")")
        return res

    def parse_optional_attribute(self) -> Optional[Attribute]:
        # Shorthand for StringAttr
        string_lit = self.parse_optional_str_literal()
        if string_lit is not None:
            return StringAttr.get(string_lit)

        # Shorthand for IntegerAttr
        integer_lit = self.parse_optional_int_literal()
        if integer_lit is not None:
            if self.peek_char(":"):
                self.parse_char(":")
                typ = self.parse_attribute()
            else:
                typ = IntegerType.get(64)
            return IntegerAttr.get(integer_lit, typ)

        # Shorthand for ArrayAttr
        parse_bracket = self.parse_optional_char("[")
        if parse_bracket:
            array = self.parse_list(self.parse_optional_attribute)
            self.parse_char("]")
            return ArrayAttr.get(array)

        # Shorthand for FlatSymvolRefAttr
        parse_at = self.parse_optional_char("@")
        if parse_at:
            symbol_name = self.parse_alpha_num(skip_white_space=False)
            return FlatSymbolRefAttr.get(symbol_name)

        parsed = self.parse_optional_char("!")
        if parsed is None:
            return None

        parsed = self.parse_optional_char("i")

        # shorthand for integer types
        if parsed:
            num = self.parse_optional_int_literal()
            if num:
                return IntegerType.get(num)
            attr_def_name = "i" + self.parse_alpha_num(skip_white_space=True)
        else:
            attr_def_name = self.parse_alpha_num(skip_white_space=True)

        attr_def = self._ctx.get_attr(attr_def_name)
        if self.parse_optional_char("<") is None:
            return attr_def()

        if issubclass(attr_def, Data):
            attr = attr_def.parse(self)
            self.parse_char(">")
            return attr

        param_list = self.parse_list(self.parse_optional_attribute)
        self.parse_char(">")
        return attr_def(param_list)

    def parse_attribute(self) -> Attribute:
        res = self.parse_optional_attribute()
        if res is None:
            raise Exception("attribute expected")
        return res

    def parse_optional_named_attribute(
            self) -> Optional[Tuple[str, Attribute]]:
        attr_name = self.parse_optional_str_literal()
        if attr_name is None:
            return None
        self.parse_char("=")
        attr = self.parse_attribute()
        return attr_name, attr

    def parse_op_attributes(self) -> Dict[str, Attribute]:
        if not self.parse_optional_char("["):
            return dict()
        attrs_with_names = self.parse_list(self.parse_optional_named_attribute)
        self.parse_char("]")
        return {name: attr for (name, attr) in attrs_with_names}

    def parse_optional_successor(self) -> Optional[Block]:
        parsed = self.parse_optional_char("^")
        if parsed is None:
            return None
        bb_name = self.parse_alpha_num(skip_white_space=False)
        if bb_name in self._blocks:
            block = self._blocks[bb_name]
            pass
        else:
            block = Block()
            self._blocks[bb_name] = block
        return block

    def parse_successors(self) -> List[Block]:
        parsed = self.parse_optional_char("(")
        if parsed is None:
            return None
        res = self.parse_list(self.parse_optional_successor, delimiter=',')
        self.parse_char(")")
        return res

    def parse_optional_op(self) -> Optional[Operation]:
        results = self.parse_optional_results()
        if results is None:
            op_name = self.parse_optional_ident()
            if op_name is None:
                return None
            results = []
        else:
            op_name = self.parse_alpha_num()

        operands = self.parse_operands()
        successors = self.parse_successors()
        attributes = self.parse_op_attributes()
        result_types = [typ for (name, typ) in results]
        op = self._ctx.get_op(op_name).create(operands,
                                              result_types,
                                              attributes=attributes,
                                              successors=successors)
        # Register the SSA value names in the parser
        for (idx, res) in enumerate(results):
            if res[0] in self._ssaValues:
                raise Exception("SSA value %s is already defined" % res[0])
            self._ssaValues[res[0]] = op.results[idx]

        region = self.parse_optional_region()
        while region is not None:
            op.add_region(region)
            region = self.parse_optional_region()
        return op

    def parse_op(self) -> Operation:
        res = self.parse_optional_op()
        if res is None:
            raise Exception("operation expected")
        return res



from xdsl.ir import *
from xdsl.irdl import *
from xdsl.util import *
from xdsl.dialects.builtin import IntegerType, Float32Type, IntegerAttr, FlatSymbolRefAttr


@dataclass
class Std:
    ctx: MLContext

    def __post_init__(self):
        self.ctx.register_op(Constant)

        self.ctx.register_op(Addi)
        self.ctx.register_op(Muli)
        self.ctx.register_op(Subi)
        self.ctx.register_op(FloordiviSigned)
        self.ctx.register_op(RemiSigned)

        self.ctx.register_op(Addf)
        self.ctx.register_op(Mulf)

        self.ctx.register_op(Call)
        self.ctx.register_op(Return)

        self.ctx.register_op(And)
        self.ctx.register_op(Or)
        self.ctx.register_op(Xor)

        self.ctx.register_op(Cmpi)

        self.f32 = Float32Type.get()
        self.i64 = IntegerType.get(64)
        self.i32 = IntegerType.get(32)
        self.i1 = IntegerType.get(1)

    # TODO make this generic in the type
    def constant(self, val: int, typ: Attribute) -> Operation:
        return Constant.create([], [typ],
                               attributes={"value": IntegerAttr.get(val, typ)})

    def constant_from_attr(self, attr: Attribute, typ: Attribute) -> Operation:
        return Constant.create([], [typ], attributes={"value": attr})

    def mulf(self, x: OpOrBlockArg, y: OpOrBlockArg) -> Operation:
        return Mulf.create([get_ssa_value(x), get_ssa_value(y)], [self.f32])

    def addf(self, x: OpOrBlockArg, y: OpOrBlockArg) -> Operation:
        return Addf.create([get_ssa_value(x), get_ssa_value(y)], [self.f32])

    def call(self, callee: str, ops: List[OpOrBlockArg],
             return_types: List[Attribute]) -> Operation:
        return Call.create(
            [get_ssa_value(op) for op in ops],
            return_types,
            attributes={"callee": FlatSymbolRefAttr.get(callee)})

    def return_(self, *ops: OpOrBlockArg) -> Operation:
        return Return.create([get_ssa_value(op) for op in ops], [])

    # TODO these operations should support all kinds of integer types
    def muli(self, x: OpOrBlockArg, y: OpOrBlockArg) -> Operation:
        return Muli.create([get_ssa_value(x), get_ssa_value(y)], [self.i32])

    def addi(self, x: OpOrBlockArg, y: OpOrBlockArg) -> Operation:
        return Addi.create([get_ssa_value(x), get_ssa_value(y)], [self.i32])

    def subi(self, x: OpOrBlockArg, y: OpOrBlockArg) -> Operation:
        return Subi.create([get_ssa_value(x), get_ssa_value(y)], [self.i32])

    def floordivi_signed(self, x: OpOrBlockArg, y: OpOrBlockArg) -> Operation:
        return FloordiviSigned.create(
            [get_ssa_value(x), get_ssa_value(y)], [self.i32])

    def remi_signed(self, x: OpOrBlockArg, y: OpOrBlockArg) -> Operation:
        return RemiSigned.create(
            [get_ssa_value(x), get_ssa_value(y)], [self.i32])

    # TODO these operations should support all kinds of integer types
    def and_(self, x: OpOrBlockArg, y: OpOrBlockArg) -> Operation:
        return And.create([get_ssa_value(x), get_ssa_value(y)], [self.i1])

    def or_(self, x: OpOrBlockArg, y: OpOrBlockArg) -> Operation:
        return Or.create([get_ssa_value(x), get_ssa_value(y)], [self.i1])

    def xor_(self, x: OpOrBlockArg, y: OpOrBlockArg) -> Operation:
        return Xor.create([get_ssa_value(x), get_ssa_value(y)], [self.i1])

    def cmpi(self, x: OpOrBlockArg, y: OpOrBlockArg, arg: int) -> Operation:
        return Cmpi.create(
            [get_ssa_value(x), get_ssa_value(y)], [self.i1],
            attributes={"predicate": IntegerAttr.get(arg, self.i64)})


@irdl_op_definition
class Constant:
    name: str = "std.constant"
    output = ResultDef(AnyAttr())
    value = AttributeDef(AnyAttr())

    # TODO verify that the output and value type are equal
    def verify_(self) -> None:
        # TODO how to force the attr to have a type? and how to query it?
        pass


@irdl_op_definition
class Addi:
    name: str = "std.addi"
    input1 = OperandDef(IntegerType)
    input2 = OperandDef(IntegerType)
    output = ResultDef(IntegerType)

    # TODO replace with trait
    def verify_(self) -> None:
        if self.input1.typ != self.input2.typ or self.input2.typ != self.output.typ:
            raise Exception("expect all input and output types to be equal")


@irdl_op_definition
class Muli:
    name: str = "std.muli"
    input1 = OperandDef(IntegerType)
    input2 = OperandDef(IntegerType)
    output = ResultDef(IntegerType)

    # TODO replace with trait
    def verify_(self) -> None:
        if self.input1.typ != self.input2.typ or self.input2.typ != self.output.typ:
            raise Exception("expect all input and output types to be equal")


@irdl_op_definition
class Subi:
    name: str = "std.subi"
    input1 = OperandDef(IntegerType)
    input2 = OperandDef(IntegerType)
    output = ResultDef(IntegerType)

    # TODO replace with trait
    def verify_(self) -> None:
        if self.input1.typ != self.input2.typ or self.input2.typ != self.output.typ:
            raise Exception("expect all input and output types to be equal")


@irdl_op_definition
class FloordiviSigned:
    name: str = "std.floordivi_signed"
    input1 = OperandDef(IntegerType)
    input2 = OperandDef(IntegerType)
    output = ResultDef(IntegerType)

    # TODO replace with trait
    def verify_(self) -> None:
        if self.input1.typ != self.input2.typ or self.input2.typ != self.output.typ:
            raise Exception("expect all input and output types to be equal")


@irdl_op_definition
class RemiSigned:
    name: str = "std.remi_signed"
    input1 = OperandDef(IntegerType)
    input2 = OperandDef(IntegerType)
    output = ResultDef(IntegerType)

    # TODO replace with trait
    def verify_(self) -> None:
        if self.input1.typ != self.input2.typ or self.input2.typ != self.output.typ:
            raise Exception("expect all input and output types to be equal")


@irdl_op_definition
class Call:
    name: str = "std.call"
    arguments = VarOperandDef(AnyAttr())
    callee = AttributeDef(FlatSymbolRefAttr)

    # Note: naming this results triggers an ArgumentError
    res = VarResultDef(AnyAttr())
    # TODO how do we verify that the types are correct?


@irdl_op_definition
class Return:
    name: str = "std.return"
    arguments = VarOperandDef(AnyAttr())


@irdl_op_definition
class And:
    name: str = "std.and"
    input1 = OperandDef(IntegerType)
    input2 = OperandDef(IntegerType)
    output = ResultDef(IntegerType)

    # TODO replace with trait
    def verify_(self) -> None:
        if self.input1.typ != self.input2.typ or self.input2.typ != self.output.typ:
            raise Exception("expect all input and output types to be equal")


@irdl_op_definition
class Or:
    name: str = "std.or"
    input1 = OperandDef(IntegerType)
    input2 = OperandDef(IntegerType)
    output = ResultDef(IntegerType)

    # TODO replace with trait
    def verify_(self) -> None:
        if self.input1.typ != self.input2.typ or self.input2.typ != self.output.typ:
            raise Exception("expect all input and output types to be equal")


@irdl_op_definition
class Xor:
    name: str = "std.xor"
    input1 = OperandDef(IntegerType)
    input2 = OperandDef(IntegerType)
    output = ResultDef(IntegerType)

    # TODO replace with trait
    def verify_(self) -> None:
        if self.input1.typ != self.input2.typ or self.input2.typ != self.output.typ:
            raise Exception("expect all input and output types to be equal")


@irdl_op_definition
class Cmpi:
    name: str = "std.cmpi"
    predicate = AttributeDef(IntegerAttr)
    input1 = OperandDef(IntegerType)
    input2 = OperandDef(IntegerType)
    output = ResultDef(IntegerType.get(1))


@irdl_op_definition
class Addf:
    name: str = "std.addf"
    input1 = OperandDef(Float32Type)
    input2 = OperandDef(Float32Type)
    output = ResultDef(Float32Type)

    # TODO replace with trait
    def verify_(self) -> None:
        if self.input1.typ != self.input2.typ or self.input2.typ != self.output.typ:
            raise Exception("expect all input and output types to be equal")


@irdl_op_definition
class Mulf:
    name: str = "std.mulf"
    input1 = OperandDef(Float32Type)
    input2 = OperandDef(Float32Type)
    output = ResultDef(Float32Type)

    # TODO replace with trait
    def verify_(self) -> None:
        if self.input1.typ != self.input2.typ or self.input2.typ != self.output.typ:
            raise Exception("expect all input and output types to be equal")



from __future__ import annotations
import inspect
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import List, Optional, Callable, Dict

from xdsl.dialects.builtin import ModuleOp
from xdsl.ir import Operation, SSAValue, Region, Block


@dataclass(eq=False, repr=False)
class RewriteAction:
    """
    Action that a single rewrite may execute.
    A rewrite always delete the matched operation, and replace it with new operations.
    The matched operation results are replaced with new ones.
    """

    new_ops: List[Operation]
    """New operations that replace the one matched."""

    new_results: List[Optional[SSAValue]]
    """SSA values that replace the matched operation results. None values are deleted SSA Values."""
    @staticmethod
    def from_op_list(new_ops: List[Operation]) -> RewriteAction:
        """
        Case where the old results will be replaced by the results of the last operation to be added.
        Can also be used with no operations (to represent deletion).
        """
        if len(new_ops) == 0:
            return RewriteAction(new_ops, [])
        else:
            return RewriteAction(new_ops, new_ops[-1].results)


class RewritePattern(ABC):
    """
    A side-effect free rewrite pattern matching on a DAG.
    """
    @abstractmethod
    def match_and_rewrite(
            self, op: Operation,
            new_operands: List[Optional[SSAValue]]) -> Optional[RewriteAction]:
        """
        Match an operation, and optionally returns a rewrite to be performed.
        `op` is the operation to match, and `new_operands` are the potential new values of the operands.
        `None` values in new_operands are deleted SSA values.
        This function returns `None` if the pattern did not match, and a rewrite action otherwise.
        """
        ...


@dataclass(eq=False, repr=False)
class AnonymousRewritePattern(RewritePattern):
    """
    A rewrite pattern encoded by an anonymous function.
    """
    func: Callable[[Operation, List[Optional[SSAValue]]],
                   Optional[RewriteAction]]

    def match_and_rewrite(
            self, op: Operation,
            new_operands: List[Optional[SSAValue]]) -> Optional[RewriteAction]:
        return self.func(op, new_operands)


def op_type_rewrite_pattern(func):
    """
    This function is intended to be used as a decorator on a RewritePatter method.
    It uses type hints to match on a specific operation type before calling the decorated function.
    """
    # Get the operation argument and check that it is a subclass of Operation
    params = [param for param in inspect.signature(func).parameters.values()]
    if len(params) < 2:
        raise Exception(
            "op_type_rewrite_pattern expects the decorated function to "
            "have two non-self arguments.")
    is_method = params[0].name == "self"
    if is_method:
        if len(params) != 3:
            raise Exception(
                "op_type_rewrite_pattern expects the decorated method to "
                "have two non-self arguments.")
    else:
        if len(params) != 2:
            raise Exception(
                "op_type_rewrite_pattern expects the decorated function to "
                "have two arguments.")
    expected_type = params[-2].annotation
    if not issubclass(expected_type, Operation):
        raise Exception(
            "op_type_rewrite_pattern expects the first non-self argument"
            "type hint to be an Operation subclass")

    if not is_method:

        def op_type_rewrite_pattern_static_wrapper(
                op: Operation,
                operands: List[Optional[SSAValue]]) -> Optional[RewriteAction]:
            if not isinstance(op, expected_type):
                return None
            return func(op, operands)

        return op_type_rewrite_pattern_static_wrapper

    def op_type_rewrite_pattern_method_wrapper(
            self, op: Operation,
            operands: List[Optional[SSAValue]]) -> Optional[RewriteAction]:
        if not isinstance(op, expected_type):
            return None
        return func(self, op, operands)

    return op_type_rewrite_pattern_method_wrapper


@dataclass(eq=False, repr=False)
class GreedyRewritePatternApplier(RewritePattern):
    """Apply a list of patterns in order until one pattern match, and then use this rewrite."""

    rewrite_patterns: List[RewritePattern]
    """The list of rewrites to apply in order."""
    def match_and_rewrite(
            self, op: Operation,
            new_operands: List[SSAValue]) -> Optional[RewriteAction]:
        for pattern in self.rewrite_patterns:
            res = pattern.match_and_rewrite(op, new_operands)
            if res is not None:
                return res

        return None


@dataclass(repr=False, eq=False)
class OperandUpdater:
    """
    Provides functionality to bookkeep changed results and to access and update them.
    """

    result_mapping: Dict[SSAValue,
                         Optional[SSAValue]] = field(default_factory=dict)
    """Map old ssa values to new values. Deleted values are mapped to None."""
    def bookkeep_results(self, old_op: Operation,
                         action: RewriteAction) -> None:
        """Bookkeep the changes made by a rewrite action matching on `old_op`."""
        if len(old_op.results) == 0:
            return

        assert len(old_op.results) == len(action.new_results)

        for (old_res, new_res) in zip(old_op.results, action.new_results):
            self.result_mapping[old_res] = new_res

    def get_new_value(self, value: SSAValue) -> Optional[SSAValue]:
        """Get the updated value, if it exists, or returns the same one."""
        return self.result_mapping.get(value, value)

    def get_new_operands(self, op: Operation) -> [Optional[SSAValue]]:
        """Get the new operation updated operands"""
        return [self.get_new_value(operand) for operand in op.operands]

    def update_operands(self, op: Operation) -> None:
        """Update an operation operands with the new operands."""
        new_operands = self.get_new_operands(op)
        if None in new_operands:
            raise Exception("Use of deleted SSA Value")
        op.operands = new_operands


@dataclass(eq=False, repr=False)
class PatternRewriteWalker:
    """
    Walks the IR in the block and instruction order, and rewrite it in place.
    Previous references to the walked operations are invalid after the walk.
    Can walk either first the regions, or first the owner operation.
    The walker will also walk recursively on the created operations.
    """

    pattern: RewritePattern
    """Pattern to apply during the walk."""

    walk_regions_first: bool = field(default=False)
    """Choose if the walker should first walk the operation regions first, or the operation itself."""

    apply_recursively: bool = field(default=True)
    """Apply recursively rewrites on new operations."""

    _updater: OperandUpdater = field(init=False,
                                     default_factory=OperandUpdater)
    """Takes care of bookkeeping the changes made during the walk."""
    def rewrite_module(self, op: ModuleOp):
        """Rewrite an entire module operation."""
        new_ops = self.rewrite_op(op)
        if len(new_ops) == 1:
            res_op = new_ops[0]
            if isinstance(res_op, ModuleOp):
                op = res_op
                return
        raise Exception(
            "Rewrite pattern did not rewrite a module into another module.")

    def rewrite_op(self, op: Operation) -> List[Operation]:
        """Rewrite an operation, along with its regions."""
        # First, we walk the regions if needed
        if self.walk_regions_first:
            self.rewrite_op_regions(op)

        # We then match for a pattern in the current operation
        action = self.pattern.match_and_rewrite(
            op, self._updater.get_new_operands(op))

        # If we produce new operations, we rewrite them recursively if requested
        if action is not None:
            self._updater.bookkeep_results(op, action)
            if self.apply_recursively:
                new_ops = []
                for new_op in action.new_ops:
                    new_ops.extend(self.rewrite_op(new_op))
                return new_ops
            else:
                for new_op in action.new_ops:
                    self._updater.update_operands(new_op)
                return action.new_ops

        # Otherwise, we update their operands, and walk recursively their regions if needed
        self._updater.update_operands(op)
        if not self.walk_regions_first:
            self.rewrite_op_regions(op)
        return [op]

    def rewrite_op_regions(self, op: Operation):
        """Rewrite the regions of an operation, and update the operation with the new regions."""
        new_regions = []
        for region in op.regions:
            new_region = Region()
            for block in region.blocks:
                new_block = Block()
                for sub_op in block.ops:
                    new_block.add_ops(self.rewrite_op(sub_op))
                new_region.add_block(new_block)
            new_regions.append(new_region)
        op.regions = []
        for region in new_regions:
            op.add_region(region)



from xdsl.ir import *
from xdsl.util import OpOrBlockArg, get_ssa_value, new_op


@dataclass
class Affine:
    ctx: MLContext

    def __post_init__(self):
        self.ctx.register_op(new_op("affine.for", 0, 0, 1))
        self.ctx.register_op(new_op("affine.load", 1, 3, 0))
        self.ctx.register_op(new_op("affine.store", 0, 4, 0))

    def for_(self, lower_bound: int, upper_bound: int,
             block: Block) -> Operation:
        op = self.ctx.get_op("affine.for").create([], [],
                                                  regions=[Region([block])])
        return op

    def load(self, value: OpOrBlockArg, i: OpOrBlockArg,
             j: OpOrBlockArg) -> Operation:
        return self.ctx.get_op("affine.load").create(
            [get_ssa_value(value),
             get_ssa_value(i),
             get_ssa_value(j)], [self.ctx.get_attr("f32")()], {})

    def store(self, value: OpOrBlockArg, place: OpOrBlockArg, i: OpOrBlockArg,
              j: OpOrBlockArg) -> Operation:
        return self.ctx.get_op("affine.store").create([
            get_ssa_value(value),
            get_ssa_value(place),
            get_ssa_value(i),
            get_ssa_value(j)
        ], [], {})



from xdsl.dialects.builtin import *
from xdsl.util import OpOrBlockArg, get_ssa_value


@dataclass
class Scf:
    ctx: MLContext

    def __post_init__(self):
        self.ctx.register_op(If)
        self.ctx.register_op(Yield)
        self.ctx.register_op(Condition)
        self.ctx.register_op(While)

    def if_(self, cond: OpOrBlockArg, true_region: Region,
            false_region: Region, return_types: List[Attribute]):
        op = If.create([get_ssa_value(cond)],
                       return_types,
                       regions=[true_region, false_region])
        return op

    def yield_(self, *ops: OpOrBlockArg):
        return Yield.create([get_ssa_value(op) for op in ops], [])

    def condition(self, cond: OpOrBlockArg, *output_ops: List[OpOrBlockArg]):
        return Condition.create([get_ssa_value(cond)] +
                                [get_ssa_value(op) for op in output_ops], [])

    def while_(self, before: Region, after: Region, ops: List[OpOrBlockArg],
               return_types: List[Attribute]):
        op = While.create([get_ssa_value(op) for op in ops],
                          return_types,
                          regions=[before, after])
        return op


@irdl_op_definition
class If:
    name: str = "scf.if"
    output = VarResultDef(AnyAttr())
    cond = OperandDef(IntegerType.get(1))

    true_region = RegionDef()
    # TODO this should be optional under certain conditions
    false_region = RegionDef()


@irdl_op_definition
class Yield:
    name: str = "scf.yield"
    arguments = VarOperandDef(AnyAttr())


@irdl_op_definition
class Condition:
    name: str = "scf.condition"
    cond = OperandDef(IntegerType.get(1))
    arguments = VarOperandDef(AnyAttr())


@irdl_op_definition
class While:
    name: str = "scf.while"
    arguments = VarOperandDef(AnyAttr())

    res = VarResultDef(AnyAttr())
    before_region = RegionDef()
    after_region = RegionDef()

    # TODO verify dependencies between scf.condition, scf.yield and the regions
    def verify_(self):
        for (idx, arg) in enumerate(self.arguments):
            if self.before_region.blocks[0].args[idx].typ != arg.typ:
                raise Exception(
                    f"Block arguments with wrong type, expected {arg.typ}, got {self.before_region.block[0].args[idx].typ}"
                )

        for (idx, res) in enumerate(self.res):
            if self.after_region.blocks[0].args[idx].typ != res.typ:
                raise Exception(
                    f"Block arguments with wrong type, expected {res.typ}, got {self.after_region.block[0].args[idx].typ}"
                )



from __future__ import annotations
from xdsl.dialects.builtin import *
from typing import TypeVar
import sys

indentNumSpaces = 2


class Printer:
    def __init__(self, stream=sys.stdout):
        self._indent: int = 0
        self._ssaValues: Dict[SSAValue, int] = dict()
        self._blockNames: Dict[Block, int] = dict()
        self._nextValidNameId: int = 0
        self._nextValidBlockId: int = 0
        self.stream = stream

    def _print(self, *args, **kwargs):
        print(*args, **kwargs, file=self.stream)

    def print_string(self, string) -> None:
        self._print(string, end='')

    T = TypeVar('T')

    def print_list(self, elems: List[T], print_fn: Callable[[T],
                                                            None]) -> None:
        if len(elems) == 0:
            return
        print_fn(elems[0])
        for elem in elems[1:]:
            self._print(", ", end='')
            print_fn(elem)

    def _print_new_line(self) -> None:
        self._print("")
        self._print(" " * self._indent * indentNumSpaces, end='')

    def _get_new_valid_name_id(self) -> int:
        self._nextValidNameId += 1
        return self._nextValidNameId - 1

    def _get_new_valid_block_id(self) -> int:
        self._nextValidBlockId += 1
        return self._nextValidBlockId - 1

    def _print_result_value(self, op: Operation, idx: int) -> None:
        val = op.results[idx]
        self._print("%", end='')
        name = self._get_new_valid_name_id()
        self._ssaValues[val] = name
        self._print("%s : " % name, end='')
        self.print_attribute(val.typ)

    def _print_results(self, op: Operation) -> None:
        results = op.results
        # No results
        if len(results) == 0:
            return

        # One result
        if len(results) == 1:
            self._print_result_value(op, 0)
            self._print(" = ", end='')
            return

        # Multiple results
        self._print("(", end='')
        self._print_result_value(op, 0)
        for idx in range(1, len(results)):
            self._print(", ", end='')
            self._print_result_value(op, idx)
        self._print(") = ", end='')

    def _print_operand(self, operand: SSAValue) -> None:
        self._print("%", end='')
        self._print("%s : " % self._ssaValues[operand], end='')
        self.print_attribute(operand.typ)

    def _print_ops(self, ops: List[Operation]) -> None:
        self._indent += 1
        for op in ops:
            self._print_new_line()
            self._print_op(op)
        self._indent -= 1
        if len(ops) > 0:
            self._print_new_line()

    def print_block_name(self, block: Block) -> None:
        self._print("^", end='')
        if block not in self._blockNames:
            self._blockNames[block] = self._get_new_valid_block_id()
        self._print(self._blockNames[block], end='')

    def _print_named_block(self, block: Block) -> None:
        self.print_block_name(block)
        if len(block.args) > 0:
            self._print("(", end='')
            self.print_list(block.args, self._print_block_arg)
            self._print(")", end='')
        self._print(": ", end='')
        self._print_ops(block.ops)

    def _print_block_arg(self, arg: BlockArgument) -> None:
        self._print("%", end='')
        name = self._get_new_valid_name_id()
        self._ssaValues[arg] = name
        self._print("%s : " % name, end='')
        self.print_attribute(arg.typ)

    def _print_region(self, region: Region) -> None:
        if len(region.blocks) == 0:
            self._print("{}", end='')
            return

        if len(region.blocks) == 1 and len(region.blocks[0].args) == 0:
            self._print(" {", end='')
            self._print_ops(region.blocks[0].ops)
            self._print("}", end='')
            return

        self._print("{", end='')
        self._print_new_line()
        for block in region.blocks:
            self._print_named_block(block)
        self._print("}", end='')

    def _print_regions(self, regions: List[Region]) -> None:
        for region in regions:
            self._print_region(region)

    def _print_operands(self, operands: List[SSAValue]) -> None:
        if len(operands) == 0:
            self._print("()", end='')
            return

        self._print("(", end='')
        self._print_operand(operands[0])
        for operand in operands[1:]:
            self._print(", ", end='')
            self._print_operand(operand)
        self._print(")", end='')

    def print_attribute(self, attribute: Attribute) -> None:
        if isinstance(attribute, IntegerType):
            width = attribute.parameters[0]
            assert isinstance(width, IntAttr)
            self._print(f'!i{width.data}', end='')
            return

        if isinstance(attribute, StringAttr):
            self._print(f'"{attribute.data}"', end='')
            return

        if isinstance(attribute, FlatSymbolRefAttr):
            self._print(f'@{attribute.parameters[0].data}', end='')
            return

        if isinstance(attribute, IntegerAttr):
            width = attribute.parameters[0]
            typ = attribute.parameters[1]
            assert (isinstance(width, IntAttr))
            self._print(width.data, end='')
            self._print(" : ", end='')
            self.print_attribute(typ)
            return

        if isinstance(attribute, ArrayAttr):
            self.print_string("[")
            self.print_list(attribute.data, self.print_attribute)
            self.print_string("]")
            return

        if isinstance(attribute, Data):
            self._print(f'!{attribute.name}<', end='')
            attribute.print(self)
            self._print(">", end='')
            return

        assert isinstance(attribute, ParametrizedAttribute)

        self._print(f'!{attribute.name}', end='')
        if len(attribute.parameters) != 0:
            self._print("<", end='')
            self.print_list(attribute.parameters, self.print_attribute)
            self._print(">", end='')

    def print_successors(self, successors: List[Block]):
        if len(successors) == 0:
            return
        self._print(" (", end='')
        self.print_list(successors, self.print_block_name)
        self._print(")", end='')

    def _print_op_attributes(self, attributes: Dict[str, Attribute]) -> None:
        if len(attributes) == 0:
            return

        self._print(" ", end='')
        self._print("[", end='')
        attribute_list = [p for p in attributes.items()]
        self._print("\"%s\" = " % attribute_list[0][0], end='')
        self.print_attribute(attribute_list[0][1])
        for (attr_name, attr) in attribute_list[1:]:
            self._print(", \"%s\" = " % attr_name, end='')
            self.print_attribute(attr)
        self._print("]", end='')

    def _print_op(self, op: Operation) -> None:
        self._print_results(op)
        self._print(op.name, end='')
        self._print_operands(op.operands)
        self.print_successors(op.successors)
        self._print_op_attributes(op.attributes)
        self._print_regions(op.regions)

    def print_op(self, op: Operation) -> None:
        self._print_op(op)
        self._print_new_line()



from __future__ import annotations
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import Dict, List, Callable, Optional, Any, TYPE_CHECKING, TypeVar
import typing

# Used for cyclic dependencies in type hints
if TYPE_CHECKING:
    from xdsl.parser import Parser
    from xdsl.printer import Printer


@dataclass
class MLContext:
    """Contains structures for operations/attributes registration."""
    _registeredOps: Dict[str,
                         typing.Type[Operation]] = field(default_factory=dict)
    _registeredAttrs: Dict[str, typing.Type[Attribute]] = field(
        default_factory=dict)

    def register_op(self, op: typing.Type[Operation]) -> None:
        """Register an operation definition. Operation names should be unique."""
        if op.name in self._registeredOps:
            raise Exception(f"Operation {op.name} has already been registered")
        self._registeredOps[op.name] = op

    def register_attr(self, attr: typing.Type[Attribute]) -> None:
        """Register an attribute definition. Attribute names should be unique."""
        if attr.name in self._registeredAttrs:
            raise Exception(
                f"Attribute {attr.name} has already been registered")
        self._registeredAttrs[attr.name] = attr

    def get_op(self, name: str) -> typing.Type[Operation]:
        if name not in self._registeredOps:
            raise Exception(f"Operation {name} is not registered")
        return self._registeredOps[name]

    def get_attr(self, name: str) -> typing.Type[Attribute]:
        if name not in self._registeredAttrs:
            raise Exception(f"Attribute {name} is not registered")
        return self._registeredAttrs[name]


@dataclass
class SSAValue(ABC):
    """A reference to an SSA variable.
    An SSA variable is either an operation result, or a basic block argument."""

    typ: Attribute
    """Each SSA variable is associated to a type."""


@dataclass
class OpResult(SSAValue):
    """A reference to an SSA variable defined by an operation result."""

    op: Operation
    """The operation defining the variable."""

    result_index: int
    """The index of the result in the defining operation."""
    def __hash__(self):
        return hash((id(self.op), self.result_index))


@dataclass
class BlockArgument(SSAValue):
    """A reference to an SSA variable defined by a basic block argument."""

    block: Block
    """The block defining the variable."""

    index: int
    """The index of the variable in the block arguments."""
    def __hash__(self):
        return hash((id(self.block), self.index))


@dataclass(frozen=True)
class Attribute(ABC):
    """
    A compile-time value.
    Attributes are used to represent SSA variable types, and can be attached
    on operations to give extra information.
    """

    name: str = field(default="", init=False)
    """The attribute name should be a static field in the attribute classes."""


@dataclass(frozen=True)
class Data(Attribute):
    """An attribute represented by a Python structure."""
    @staticmethod
    @abstractmethod
    def parse(parser: Parser) -> Data:
        """Parse the attribute value."""
        ...

    @abstractmethod
    def print(self, printer: Printer) -> None:
        """Print the attribute value."""
        ...


@dataclass(frozen=True)
class ParametrizedAttribute(Attribute):
    """An attribute parametrized by other attributes."""

    name: str = field(default="", init=False)
    parameters: List[Attribute] = field(default_factory=list)

    def __post_init__(self):
        self.verify()

    def verify(self) -> None:
        ...


T = TypeVar('T', bound='Operation')


@dataclass
class Operation(ABC):
    """A generic operation. Operation definitions inherit this class."""

    name: str = field(default="", init=False)
    """The operation name. Should be a static member of the class"""

    operands: List[SSAValue] = field(default_factory=list)
    """The operation operands."""

    results: List[OpResult] = field(default_factory=list)
    """The results created by the operation."""

    successors: List[Block] = field(default_factory=list)
    """
    The basic blocks that the operation may give control to.
    This list should be empty for non-terminator operations.
    """

    attributes: Dict[str, Attribute] = field(default_factory=dict)
    """The attributes attached to the operation."""

    regions: List[Region] = field(default_factory=list)
    """Regions arguments of the operation."""

    parent: Optional[Block] = field(default=None)
    """The block containing this operation."""
    def __post_init__(self):
        assert (self.name != "")
        assert (isinstance(self.name, str))

    @staticmethod
    def with_result_types(op: Any,
                          operands: Optional[List[SSAValue]] = None,
                          result_types: Optional[List[Attribute]] = None,
                          attributes: Optional[Dict[str, Attribute]] = None,
                          successors: Optional[List[Block]] = None,
                          regions: Optional[List[Region]] = None) -> Operation:
        operation = op()
        if operands is not None:
            operation.operands = operands
        if result_types is not None:
            operation.results = [
                OpResult(typ, operation, idx)
                for (idx, typ) in enumerate(result_types)
            ]
        if attributes is not None:
            operation.attributes = attributes
        if successors is not None:
            operation.successors = successors
        if regions is not None:
            for region in regions:
                operation.add_region(region)
        return operation

    @classmethod
    def create(cls: typing.Type[T],
               operands: Optional[List[SSAValue]] = None,
               result_types: Optional[List[Attribute]] = None,
               attributes: Optional[Dict[str, Attribute]] = None,
               successors: Optional[List[Block]] = None,
               regions: Optional[List[Region]] = None) -> T:
        return Operation.with_result_types(cls, operands, result_types,
                                           attributes, successors, regions)

    def add_region(self, region: Region) -> None:
        self.regions.append(region)
        region.parent = self

    def walk(self, fun: Callable[[Operation], None]) -> None:
        """Call a function on all operations contained in the operation (including this one)"""
        fun(self)
        for region in self.regions:
            region.walk(fun)

    def verify(self) -> None:
        self.verify_()
        for region in self.regions:
            region.verify()

    @abstractmethod
    def verify_(self) -> None:
        ...

    def __eq__(self, other: Operation) -> bool:
        return self is other

    def __hash__(self) -> int:
        return id(self)


@dataclass(eq=False)
class Block:
    """A sequence of operations"""

    args: List[BlockArgument] = field(default_factory=list)
    """The basic block arguments."""

    ops: List[Operation] = field(default_factory=list)
    """Ordered operations contained in the block."""

    parent: Optional[Region] = field(default=None)
    """Parent region containing the block."""
    @staticmethod
    def with_arg_types(arg_types: List[Attribute]) -> Block:
        b = Block()
        b.args = [
            BlockArgument(typ, b, index) for index, typ in enumerate(arg_types)
        ]
        return b

    def add_op(self, operation: Operation) -> None:
        self.ops.append(operation)
        operation.parent = self

    def add_ops(self, ops: List[Operation]) -> None:
        for op in ops:
            self.add_op(op)

    def walk(self, fun: Callable[[Operation], None]) -> None:
        """Call a function on all operations contained in the block."""
        for op in self.ops:
            op.walk(fun)

    def verify(self) -> None:
        for operation in self.ops:
            operation.verify()

    def __eq__(self, other: Block) -> bool:
        return self is other

    def __hash__(self) -> int:
        return id(self)


@dataclass
class Region:
    """A region contains a CFG of blocks. Regions are contained in operations."""

    blocks: List[Block] = field(default_factory=list)
    """Blocks contained in the region. The first block is the entry block."""

    parent: Optional[Operation] = field(default=None)
    """Operation containing the region."""
    @staticmethod
    def from_operation_list(ops: List[Operation]) -> Region:
        block = Block([], ops)
        region = Region()
        region.add_block(block)
        return region

    def add_block(self, block: Block) -> None:
        self.blocks.append(block)
        block.parent = self

    def walk(self, fun: Callable[[Operation], None]) -> None:
        """Call a function on all operations contained in the region."""
        for block in self.blocks:
            block.walk(fun)

    def verify(self) -> None:
        for block in self.blocks:
            block.verify()



from inspect import signature
from typing import Union, NoReturn, Callable, List, overload
import typing
from dataclasses import dataclass
from xdsl.dialects.builtin import FuncOp, FunctionType, ModuleOp, StringAttr
from xdsl.ir import Operation, SSAValue, BlockArgument, Block, Region, Attribute


def assert_never(value: NoReturn) -> NoReturn:
    assert False, f'Unhandled value: {value} ({type(value).__name__})'


OpOrBlockArg = Union[Operation, BlockArgument]


def get_ssa_value(x: OpOrBlockArg) -> SSAValue:
    if isinstance(x, Operation):
        return x.results[0]
    elif isinstance(x, BlockArgument):
        return x
    else:
        assert_never(x)


def new_op(op_name, num_results, num_operands,
           num_regions) -> typing.Type[Operation]:
    @dataclass(eq=False)
    class OpBase(Operation):
        name: str = op_name

        def verify_(self) -> None:
            if len(self.results) != num_results or len(
                    self.operands) != num_operands or len(
                        self.regions) != num_regions:
                raise Exception("%s verifier" % op_name)

    return OpBase


def new_type(type_name):
    @dataclass(frozen=True, eq=False)
    class TypeBase(Attribute):
        name: str = type_name

        def verify_(self) -> None:
            if len(self.parameters) != 0:
                raise Exception(f"{type_name} should have no parameters")

    return TypeBase


@overload
def block(t: Attribute, f: Callable[[BlockArgument],
                                    List[Operation]]) -> Block:
    ...


@overload
def block(
        ts: List[Attribute], f: Callable[[BlockArgument, BlockArgument],
                                         List[Operation]]) -> Block:
    ...


@overload
def block(
    ts: List[Attribute],
    f: Callable[[BlockArgument, BlockArgument, BlockArgument], List[Operation]]
) -> Block:
    ...


def block(t, f) -> Block:
    if isinstance(t, Attribute) and callable(f) and len(
            signature(f).parameters) == 1:
        b = Block.with_arg_types([t])
        b.add_ops(f(b.args[0]))
        return b
    if isinstance(t, list) and callable(f) and len(
            signature(f).parameters) == len(t):
        b = Block.with_arg_types(t)
        b.add_ops(f(*b.args))
        return b


@overload
def func(name: str, type: Attribute, return_type: Attribute,
         f: Callable[[BlockArgument], List[Operation]]) -> Operation:
    ...


@overload
def func(
        name: str, types: List[Attribute], return_type: Attribute,
        f: Callable[[BlockArgument, BlockArgument],
                    List[Operation]]) -> Operation:
    ...


@overload
def func(
    name: str, types: List[Attribute], return_type: Attribute,
    f: Callable[[BlockArgument, BlockArgument, BlockArgument], List[Operation]]
) -> Operation:
    ...


def func(name, input_types, return_types, f) -> Operation:
    type_attr = FunctionType.get(input_types, return_types)
    op = FuncOp.create(
        [], [],
        attributes={
            "sym_name": StringAttr.get(name),
            "type": type_attr,
            "sym_visibility": StringAttr.get("private")
        },
        regions=[Region([block(input_types, f)])])
    return op


# This function is easier to use while parsing, but makes the
# inline definitions complicated
def func2(name, input_types, return_types, region: Region) -> Operation:
    type_attr = FunctionType.get(input_types, return_types)
    op = FuncOp.create(
        [], [],
        attributes={
            "sym_name": StringAttr.get(name),
            "type": type_attr,
            "sym_visibility": StringAttr.get("private")
        },
        regions=[region])
    return op


def module(ops: Union[List[Operation], Region]) -> Operation:
    if isinstance(ops, List):
        region = Region([Block([], ops)])
    else:
        region = ops
    op = ModuleOp.create([], [], regions=[region])
    return op



#!/usr/bin/env python3

import argparse
import sys
from xdsl.ir import *
from xdsl.parser import *
from xdsl.printer import *
from xdsl.dialects.std import *
from xdsl.dialects.scf import *
from xdsl.dialects.affine import *
from xdsl.dialects.builtin import *

arg_parser = argparse.ArgumentParser(
    description='MLIR modular optimizer driver')
arg_parser.add_argument("-f",
                        type=str,
                        required=False,
                        help="path to input file")


def __main__(input_str: str):
    ctx = MLContext()
    builtin = Builtin(ctx)
    std = Std(ctx)
    affine = Affine(ctx)
    scf = Scf(ctx)

    parser = Parser(ctx, input_str)
    module = parser.parse_op()
    module.verify()

    printer = Printer()
    printer.print_op(module)


def main():
    args = arg_parser.parse_args()
    if not args.f:
        input_str = sys.stdin.read()
    else:
        f = open(args.f, mode='r')
        input_str = f.read()

    __main__(input_str)


if __name__ == "__main__":
    main()



import mlir.ir
import array
from tests.affine_test import *
from xdsl.dialects.scf import *
from xdsl.dialects.std import *
from xdsl.dialects.builtin import *
from xdsl.dialects.memref import *
from xdsl.dialects.affine import *
from xdsl.parser import *

test_prog = """
module() {
builtin.func() ["sym_name" = "test", "type" = !fun<[], []>, "sym_visibility" = "private"]
{
  memref.global() ["sym_name" = !symbol_name<"g">, "type" = !i32, "initial_value" = 0 : !i32, "sym_visibility" = "public"]
  %0 : !memref<[1 : !i64], !i32> = memref.get_global() ["name" = @g]
  %1 : !memref<[1 : !i64], !i32> = memref.alloca() ["alignment" = 0 : !i64, "operand_segment_sizes" = !vector<[0 : !i64, 0 : !i64]>]
  %2 : !i32 = std.constant() ["value" = 42 : !i32]
  memref.store(%2 : !i32, %1 : !memref<[1 : !i64], !i32>) 
  %3 : !i32 = memref.load(%1 : !memref<[1 : !i64], !i32>) 
  %4 : !memref<[10 : !i64, 2 : !i64], !i32> = memref.alloc() ["alignment" = 0 : !i64, "operand_segment_sizes" = !vector<[0 : !i64, 0 : !i64]>]
  memref.store(%2 : !i32, %4 : !memref<[10 : !i64, 2 : !i64], !i32>, %2 : !i32, %3 : !i32
) 
  memref.dealloc(%1 : !memref<[1 : !i64], !i32>) 
  memref.dealloc(%4 : !memref<[10 : !i64, 2 : !i64], !i32>) 
}
}
"""

#memref.store(%2 : !i32, %4 : !memref<[10 : !i64, 2 : !i64], !i32>, %2 : !i32, %3 : !i32)
#memref.dealloc(%4 : !memref<[10 : !i64, 2 : !i64], !i32>)


class MLIRConverter:
    def __init__(self, ctx):
        self.ctx = ctx
        self.op_to_mlir_ops: Dict[Operation, mlir.ir.Operation] = dict()
        self.block_to_mlir_blocks: Dict[Block, mlir.ir.Block] = dict()

    def convert_function_type(self, typ: FunctionType) -> mlir.ir.FunctionType:
        input_array = typ.parameters[0]
        output_array = typ.parameters[1]
        inputs = [
            self.convert_type(input_type_attr)
            for input_type_attr in input_array.data
        ]
        outputs = [
            self.convert_type(output_type_attr)
            for output_type_attr in output_array.data
        ]
        return mlir.ir.FunctionType.get(inputs, outputs)

    def convert_type(self, typ: Attribute) -> mlir.ir.Type:
        if isinstance(typ, Float32Type):
            return mlir.ir.F32Type.get()
        if isinstance(typ, IntegerType):
            return mlir.ir.IntegerType.get_signless(typ.width.data)
        if isinstance(typ, IndexType):
            return mlir.ir.IndexType.get()
        if isinstance(typ, FunctionType):
            return self.convert_function_type(typ)
        if isinstance(typ, MemRefType):
            return mlir.ir.MemRefType.get(typ.get_shape(),
                                          self.convert_type(typ.element_type))
        raise Exception(f"Unsupported type for mlir conversion: {typ}")

    def convert_value(self, value: SSAValue) -> mlir.ir.Value:
        if isinstance(value, OpResult):
            mlir_op = self.op_to_mlir_ops[value.op]
            return mlir_op.results[value.result_index]
        elif isinstance(value, BlockArgument):
            mlir_block = self.block_to_mlir_blocks[value.block]
            return mlir_block.arguments[value.index]
        raise Exception("Unknown value")

    def convert_attribute(self, attr: Attribute) -> mlir.ir.Attribute:
        if isinstance(attr, StringAttr):
            return mlir.ir.StringAttr.get(attr.data)
        if isinstance(attr, IntegerAttr):
            return mlir.ir.IntegerAttr.get(
                self.convert_type(attr.parameters[1]), attr.parameters[0].data)
        if isinstance(attr, ArrayAttr):
            return mlir.ir.ArrayAttr.get(
                [self.convert_attribute(sub_attr) for sub_attr in attr.data])
        if isinstance(attr, VectorAttr):
            # TODO ugliest hack I ever wrote
            return mlir.ir.DenseIntElementsAttr.parse(
                f"dense<{[d.parameters[0].data for d in attr.data.data]}> : vector<{len(attr.data.data)}xi32>"
            )
        if isinstance(attr, FlatSymbolRefAttr):
            return mlir.ir.FlatSymbolRefAttr.get(attr.parameters[0].data)
        # SymbolNameAttrs are in fact just StringAttrs
        if isinstance(attr, SymbolNameAttr):
            return mlir.ir.StringAttr.get(attr.parameters[0].data)
        try:
            return mlir.ir.TypeAttr.get(self.convert_type(attr))
        except Exception:
            raise Exception(
                f"Unsupported attribute for mlir conversion: {attr}")

    def convert_op(self, op: Operation) -> mlir.ir.Operation:
        result_types = [self.convert_type(result.typ) for result in op.results]
        operands = [self.convert_value(operand) for operand in op.operands]
        attributes = {
            name: self.convert_attribute(attr)
            for name, attr in op.attributes.items()
        }

        mlir_op = mlir.ir.Operation.create(op.name,
                                           results=result_types,
                                           operands=operands,
                                           attributes=attributes,
                                           regions=len(op.regions))
        self.op_to_mlir_ops[op] = mlir_op

        for region_idx in range(len(op.regions)):
            self.convert_region(op.regions[region_idx],
                                mlir_op.regions[region_idx])
        return mlir_op

    def convert_block(self, block: Block, mlir_block: mlir.ir.Block) -> None:
        ip = mlir.ir.InsertionPoint.at_block_begin(mlir_block)
        for op in block.ops:
            ip.insert(self.convert_op(op))
        return mlir_block

    def convert_region(self, region: Region,
                       mlir_region: mlir.ir.Region) -> None:
        assert (len(mlir_region.blocks) == 0)
        for block in region.blocks:
            mlir_block_args = [
                self.convert_type(block_operand.typ)
                for block_operand in block.args
            ]
            mlir_block = mlir_region.blocks.append(*mlir_block_args)
            self.block_to_mlir_blocks[block] = mlir_block
            self.convert_block(block, mlir_block)

    def convert_module(self, op: Operation) -> mlir.ir.Module:
        with mlir.ir.Context() as mlir_ctx:
            mlir_ctx.allow_unregistered_dialects = True
            with mlir.ir.Location.unknown(mlir_ctx):
                if not isinstance(op, ModuleOp):
                    raise Exception("top-level operation should be a ModuleOp")
                mlir_module = mlir.ir.Module.create()
                mlir_block = mlir_module.operation.regions[0].blocks[0]
                block = op.regions[0].blocks[0]

                ip = mlir.ir.InsertionPoint.at_block_begin(mlir_block)
                for op in block.ops:
                    ip.insert(self.convert_op(op))
                return mlir_module


def __main__():
    ctx = MLContext()
    builtin = Builtin(ctx)
    std = Std(ctx)
    affine = Affine(ctx)
    scf = Scf(ctx)
    memref = MemRef(ctx)

    parser = Parser(ctx, test_prog)
    module = parser.parse_op()
    module.verify()

    converter = MLIRConverter(ctx)
    mlir_module = converter.convert_module(module)
    print(mlir_module)

    affine_func = get_example_affine_program(ctx, builtin, std, affine)
    with mlir.ir.Context() as mlir_ctx:
        mlir_ctx.allow_unregistered_dialects = True
        with mlir.ir.Location.unknown(mlir_ctx):
            mlir_affine_func = converter.convert_op(affine_func)
            print(mlir_affine_func)


if __name__ == "__main__":
    __main__()



from dataclasses import dataclass
from xdsl.ir import *
from xdsl.util import OpOrBlockArg, get_ssa_value
from xdsl.dialects.builtin import *
from xdsl.irdl import *


@dataclass
class MemRef:
    ctx: MLContext

    def __post_init__(self):
        self.ctx.register_attr(MemRefType)

        self.ctx.register_op(Load)
        self.ctx.register_op(Store)
        self.ctx.register_op(Alloc)
        self.ctx.register_op(Alloca)
        self.ctx.register_op(Dealloc)

        self.ctx.register_op(GetGlobal)
        self.ctx.register_op(Global)

    def load(self,
             value: OpOrBlockArg,
             indices: List[OpOrBlockArg] = []) -> Operation:
        # TODO should we really check these things here?
        if not isinstance(get_ssa_value(value).typ, MemRefType):
            raise Exception("memref.load expected a MemRefType operand")
        return Load.create([get_ssa_value(value)] +
                           [get_ssa_value(i) for i in indices],
                           [get_ssa_value(value).typ.element_type])

    def store(self,
              value: OpOrBlockArg,
              place: OpOrBlockArg,
              indices: List[OpOrBlockArg] = []) -> Operation:
        return Store.create(
            [get_ssa_value(value), get_ssa_value(place)] +
            [get_ssa_value(i) for i in indices], [])

    def alloc(self,
              alignment: int,
              return_type: Attribute,
              shape: List[int] = [1]) -> Operation:
        return Alloc.create(
            [], [MemRefType.get(return_type, shape)],
            attributes={
                "alignment": IntegerAttr.get(alignment, IntegerType.get(64)),
                "operand_segment_sizes": VectorAttr.get([0, 0])
            })

    def alloca(self,
               alignment: int,
               return_type: Attribute,
               shape: List[int] = [1]) -> Operation:
        return Alloca.create(
            [], [MemRefType.get(return_type, shape)],
            attributes={
                "alignment": IntegerAttr.get(alignment, IntegerType.get(64)),
                "operand_segment_sizes": VectorAttr.get([0, 0])
            })

    def dealloc(self, memref: OpOrBlockArg) -> Operation:
        return Dealloc.create([get_ssa_value(memref)], [])

    def get_global(self, name: str, return_type: Attribute) -> Operation:
        return GetGlobal.create([], [return_type],
                                {"name": FlatSymbolRefAttr.get(name)})

    def global_(self,
                sym_name: str,
                typ: Attribute,
                initial_value: Optional[Attribute],
                sym_visibility: str = "private",
                constant: bool = False) -> Operation:
        if not initial_value:
            raise Exception("optional arguments are not yet supported")
        return Global.create(
            [], [], {
                "sym_name": SymbolNameAttr.get(sym_name),
                "type": typ,
                "initial_value": initial_value,
                "sym_visibility": StringAttr.get(sym_visibility)
            })


@irdl_attr_definition
class MemRefType:
    name = "memref"

    shape = ParameterDef(ArrayOfConstraint(IntegerAttr))
    element_type = ParameterDef(AnyAttr())

    def get_num_dims(self) -> int:
        return len(self.parameters[0].data)

    def get_shape(self) -> List[int]:
        return [i.parameters[0].data for i in self.shape.data]

    @staticmethod
    def get(referenced_type: Attribute, shape: List[int] = [1]) -> Attribute:
        return MemRefType([
            ArrayAttr.get(
                [IntegerAttr.get(d, IntegerType.get(64)) for d in shape]),
            referenced_type
        ])


@irdl_op_definition
class Load:
    name = "memref.load"
    memref = OperandDef(MemRefType)
    indices = VarOperandDef(IndexType)
    res = ResultDef(AnyAttr())

    # TODO varargs for indexing, which must match the memref dimensions
    # Problem: memref dimensions require variadic type parameters, which is subject to change

    def verify_(self):
        if self.memref.typ.element_type != self.res.typ:
            raise Exception(
                "expected return type to match the MemRef element type")

        if self.memref.typ.get_num_dims() != len(self.indices):
            raise Exception("expected an index for each dimension")


@irdl_op_definition
class Store:
    name = "memref.store"
    value = OperandDef(AnyAttr())
    memref = OperandDef(MemRefType)
    indices = VarOperandDef(IndexType)

    def verify_(self):
        if self.memref.typ.element_type != self.value.typ:
            raise Exception(
                "expected value type to match the MemRef element type")

        if self.memref.typ.get_num_dims() != len(self.indices):
            raise Exception("expected an index for each dimension")


@irdl_op_definition
class Alloc:
    name = "memref.alloc"

    dynamic_sizes = VarOperandDef(IndexType)
    symbol_operands = VarOperandDef(IndexType)

    memref = ResultDef(MemRefType)

    # TODO how to constraint the IntegerAttr type?
    alignment = AttributeDef(IntegerAttr)

    irdl_options = [AttrSizedOperandSegments()]


@irdl_op_definition
class Alloca:
    name = "memref.alloca"

    dynamic_sizes = VarOperandDef(IndexType)
    symbol_operands = VarOperandDef(IndexType)

    memref = ResultDef(MemRefType)

    # TODO how to constraint the IntegerAttr type?
    alignment = AttributeDef(IntegerAttr)

    irdl_options = [AttrSizedOperandSegments()]


@irdl_op_definition
class Dealloc:
    name = "memref.dealloc"
    memref = OperandDef(MemRefType)


#@irdl_op_definition
class GetGlobal(Operation):
    name = "memref.get_global"
    #name = AttributeDef(FlatSymbolRefAttr)

    memref = ResultDef(MemRefType)

    def verify_(self) -> None:
        if not 'name' in self.attributes:
            raise Exception("GetGlobal requires a 'name' attribute")

        if not isinstance(self.attributes['name'], FlatSymbolRefAttr):
            raise Exception(
                "expected 'name' attribute to be a FlatSymbolRefAttr")

    #TODO how to verify the types, as the global might be defined in another compilation unit


@irdl_op_definition
class Global:
    name = "memref.global"
    sym_name = AttributeDef(SymbolNameAttr)
    sym_visibility = AttributeDef(StringAttr)
    type = AttributeDef(AnyAttr())

    # TODO should be optional
    initial_value = AttributeDef(AnyAttr())

    # TODO how do we represent these in MLIR-Lite
    #constant = AttributeDef(UnitAttr)



from __future__ import annotations
from dataclasses import dataclass
from abc import ABC, abstractmethod

from xdsl.ir import Operation, Attribute, ParametrizedAttribute, SSAValue
from typing import List, Tuple, Optional, Union
from inspect import isclass
import typing


@dataclass
class AttrConstraint(ABC):
    """Constrain an attribute to a certain value."""
    @abstractmethod
    def verify(self, attr: Attribute) -> None:
        """
        Check if the attribute satisfies the constraint,
        or raise an exception otherwise.
        """
        ...


@dataclass
class EqAttrConstraint(AttrConstraint):
    """Constrain an attribute to be equal to another attribute."""

    attr: Attribute
    """The attribute we want to check equality with."""
    def verify(self, attr: Attribute) -> None:
        if attr != self.attr:
            raise Exception(f"Expected attribute {self.attr} but got {attr}")


@dataclass
class BaseAttr(AttrConstraint):
    """Constrain an attribute to be of a given base type."""

    attr: typing.Type[Attribute]
    """The expected attribute base type."""
    def verify(self, attr: Attribute) -> None:
        if not isinstance(attr, self.attr):
            raise Exception(
                f"{attr} should be of base attribute {self.attr.name}")


def attr_constr_coercion(
    attr: Union[Attribute, typing.Type[Attribute], AttrConstraint]
) -> AttrConstraint:
    """
    Attributes are coerced into EqAttrConstraints,
    and Attribute types are coerced into BaseAttr.
    """
    if isinstance(attr, Attribute):
        return EqAttrConstraint(attr)
    if isclass(attr) and issubclass(attr, Attribute):
        return BaseAttr(attr)
    return attr


@dataclass
class AnyAttr(AttrConstraint):
    """Constraint that is verified by all attributes."""
    def verify(self, attr: Attribute) -> None:
        pass


@dataclass(init=False)
class AnyOf(AttrConstraint):
    """Ensure that an attribute satisfies one of the given constraints."""

    attr_constrs: List[AttrConstraint]
    """The list of constraints that are checked."""
    def __init__(self,
                 attr_constrs: List[Union[Attribute, typing.Type[Attribute],
                                          AttrConstraint]]):
        self.attr_constrs = [
            attr_constr_coercion(constr) for constr in attr_constrs
        ]

    def verify(self, attr: Attribute) -> None:
        for attr_constr in self.attr_constrs:
            if isinstance(attr_constr, Attribute) and attr_constr == attr:
                return
            try:
                attr_constr.verify(attr)
                return
            except Exception:
                pass
        raise Exception(f"Unexpected attribute {attr}")


@dataclass(init=False)
class ParamAttrConstraint(AttrConstraint):
    """
    Constrain an attribute to be of a given type,
    and also constrain its parameters with additional constraints.
    """

    base_attr: typing.Type[Attribute]
    """The base attribute type."""

    param_constrs: List[AttrConstraint]
    """The attribute parameter constraints"""
    def __init__(self, base_attr: typing.Type[Attribute],
                 param_constrs: List[Union[Attribute, typing.Type[Attribute],
                                           AttrConstraint]]):
        self.base_attr = base_attr
        self.param_constrs = [
            attr_constr_coercion(constr) for constr in param_constrs
        ]

    def verify(self, attr: Attribute) -> None:
        assert isinstance(attr, ParametrizedAttribute)
        if not isinstance(attr, self.base_attr):
            raise Exception(
                f"Base attribute {self.base_attr.name} expected, but got {attr.name}"
            )
        if len(self.param_constrs) != len(attr.parameters):
            raise Exception(
                f"{len(self.param_constrs)} parameters expected, but got {len(attr.parameters)}"
            )
        for idx, param_constr in enumerate(self.param_constrs):
            param_constr.verify(attr.parameters[idx])


@dataclass
class IRDLOption(ABC):
    """Additional option used in IRDL."""
    ...


@dataclass
class AttrSizedOperandSegments(IRDLOption):
    """Expect an attribute on the op that contains the sizes of the variadic operands."""

    attribute_name = "operand_segment_sizes"
    """Name of the attribute containing the variadic operand sizes."""


@dataclass
class AttrSizedResultSegments(IRDLOption):
    """Expect an attribute on the op that contains the sizes of the variadic results."""

    attribute_name = "result_segment_sizes"
    """Name of the attribute containing the variadic result sizes."""


@dataclass
class OperandOrResultDef(ABC):
    """An operand or a result definition. Should not be used directly."""
    ...


@dataclass
class VariadicDef(OperandOrResultDef):
    """A variadic operand or result definition. Should not be used directly."""
    ...


@dataclass
class OptionalDef(OperandOrResultDef):
    """An optional operand or result definition. Should not be used directly."""
    ...


@dataclass(init=False)
class OperandDef(OperandOrResultDef):
    """An IRDL operand definition."""

    constr: AttrConstraint
    """The operand constraint."""
    def __init__(self, typ: Union[Attribute, typing.Type[Attribute],
                                  AttrConstraint]):
        self.constr = attr_constr_coercion(typ)


@dataclass(init=False)
class VarOperandDef(OperandDef, VariadicDef):
    """An IRDL variadic operand definition."""


@dataclass(init=False)
class OptOperandDef(VarOperandDef, OptionalDef):
    """An IRDL optional operand definition."""


@dataclass(init=False)
class ResultDef(OperandOrResultDef):
    """An IRDL result definition."""

    constr: AttrConstraint
    """The result constraint."""
    def __init__(self, typ: Union[Attribute, typing.Type[Attribute],
                                  AttrConstraint]):
        self.constr = attr_constr_coercion(typ)


@dataclass(init=False)
class VarResultDef(ResultDef, VariadicDef):
    """An IRDL variadic result definition."""


@dataclass(init=False)
class OptResultDef(VarResultDef, OptionalDef):
    """An IRDL optional result definition."""


@dataclass
class RegionDef:
    """
    An IRDL region definition.
    If the block_args is specified, then the region expect to have the entry block with these arguments.
    """
    block_args: Optional[List[Attribute]] = None


@dataclass
class SingleBlockRegionDef(RegionDef):
    """An IRDL region definition that expects exactly one block."""
    pass


@dataclass(init=False)
class AttributeDef:
    """An IRDL attribute definition."""

    constr: AttrConstraint
    """The attribute constraint."""
    def __init__(self, typ: Union[Attribute, typing.Type[Attribute],
                                  AttrConstraint]):
        self.constr = attr_constr_coercion(typ)


def get_variadic_sizes(op: Operation, is_operand: bool) -> List[int]:
    """Get variadic sizes of operands or results."""

    # We need irdl to define VectorAttr, but here we need VectorAttr.
    # So we have a circular dependency that we solve by importing in this function.
    from xdsl.dialects.builtin import VectorAttr

    operand_or_result_defs = op.irdl_operand_defs if is_operand else op.irdl_result_defs
    variadic_defs = [(arg_name, arg_def)
                     for arg_name, arg_def in operand_or_result_defs
                     if isinstance(arg_def, VariadicDef)]
    options = op.irdl_options

    op_defs = op.operands if is_operand else op.results
    def_type_name = "operand" if is_operand else "result"

    # If the size is in the attributes, fetch it
    attribute_option = AttrSizedOperandSegments(
    ) if is_operand else AttrSizedResultSegments()
    if attribute_option in options:
        size_attribute_name = AttrSizedOperandSegments.attribute_name if is_operand else AttrSizedResultSegments.attribute_name
        if size_attribute_name not in op.attributes:
            raise Exception(
                f"Expected {size_attribute_name} attribute in {op.name} operation."
            )
        attribute = op.attributes[size_attribute_name]
        if not isinstance(attribute, VectorAttr):
            raise Exception(
                f"{size_attribute_name} attribute is expected to be a VectorAttr."
            )
        variadic_sizes = [
            size_attr.value.data for size_attr in attribute.data.data
        ]
        if len(variadic_sizes) != len(variadic_defs):
            raise Exception(
                f"expected {len(variadic_defs)} values in {size_attribute_name}, but got {len(variadic_sizes)}"
            )
        if len(operand_or_result_defs) - len(variadic_defs) + sum(
                variadic_sizes) != len(op_defs):
            raise Exception(
                f"{size_attribute_name} values does not correspond to variadic arguments sizes."
            )
        return variadic_sizes

    # If there are no variadics arguments, we just check that we have the right number of arguments
    if len(variadic_defs) == 0:
        if len(op_defs) != len(operand_or_result_defs):
            raise Exception(
                f"Expected {len(operand_or_result_defs)} {'operands' if is_operand else 'results'}, but got {len(op_defs)}"
            )
        return []

    # If there is a single variadic argument, we can get its size from the number of arguments.
    if len(variadic_defs) == 1:
        if len(op_defs) - len(operand_or_result_defs) + 1 < 0:
            raise Exception(
                f"Expected at least {len(operand_or_result_defs) - 1} {def_type_name}s, got {len(operand_or_result_defs)}"
            )
        return [len(op_defs) - len(operand_or_result_defs) + 1]

    # Unreachable, all cases should have been handled.
    # Additional cases should raise an exception upon definition of the irdl operation.
    assert False


def get_operand_or_result(
        op: Operation, arg_def_idx: int, previous_var_args: int,
        is_operand: bool
) -> Union[SSAValue, Optional[SSAValue], List[SSAValue]]:
    """
    Get an operand or a result.
    In the case of a variadic operand or result definition, return a list of operand or results.
    :param op: The operation we want to get the operand or result of.
    :param arg_def_idx: The operand or result index in the irdl definition.
    :param previous_var_args: The number of previous variadic operands or results definition before this definition.
    :param is_operand: Do we get the operand or the result.
    :return:
    """
    argument_defs = op.irdl_operand_defs if is_operand else op.irdl_result_defs
    op_arguments = op.operands if is_operand else op.results

    variadic_sizes = get_variadic_sizes(op, is_operand)

    begin_arg = arg_def_idx - previous_var_args + sum(
        variadic_sizes[:previous_var_args])
    if isinstance(argument_defs[arg_def_idx][1], OptionalDef):
        arg_size = variadic_sizes[previous_var_args]
        if arg_size == 0:
            return None
        else:
            return op_arguments[begin_arg]
    if isinstance(argument_defs[arg_def_idx][1], VariadicDef):
        arg_size = variadic_sizes[previous_var_args]
        return op_arguments[begin_arg:begin_arg + arg_size]
    else:
        return op_arguments[begin_arg]


def irdl_op_verify(op: Operation, operands: List[Tuple[str, OperandDef]],
                   results: List[Tuple[str, ResultDef]],
                   regions: List[Tuple[str, RegionDef]],
                   attributes: List[Tuple[str, AttributeDef]]) -> None:
    """Given an IRDL definition, verify that an operation satisfies its invariants."""

    # Verify operands.
    # get_variadic_sizes already verify that the variadic operand sizes match the number of operands.
    operand_sizes = get_variadic_sizes(op, is_operand=True)
    current_operand = 0
    current_var_operand = 0
    for operand_name, operand_def in operands:
        if isinstance(operand_def, VarOperandDef):
            for i in range(operand_sizes[current_var_operand]):
                operand_def.constr.verify(op.operands[current_operand].typ)
                current_operand += 1
            current_operand += 1
        else:
            operand_def.constr.verify(op.operands[current_operand].typ)
            current_operand += 1

    # Verify results
    # get_variadic_sizes already verify that the variadic result sizes match the number of results.
    result_sizes = get_variadic_sizes(op, is_operand=False)
    current_result = 0
    current_var_result = 0
    for result_name, result_def in results:
        if isinstance(result_def, VarResultDef):
            for i in range(result_sizes[current_var_result]):
                result_def.constr.verify(op.results[current_result].typ)
                current_result += 1
            current_result += 1
        else:
            result_def.constr.verify(op.results[current_result].typ)
            current_result += 1

    if len(regions) != len(op.regions):
        raise Exception(
            f"op has {len(op.regions)} regions, but {len(regions)} were expected"
        )

    for idx, (region_name, region_def) in enumerate(regions):
        if isinstance(
                region_def,
                SingleBlockRegionDef) and len(op.regions[idx].blocks) != 1:
            raise Exception(
                f"region {region_name} at position {idx} should have a single block, but got {len(op.regions[idx].blocks)} blocks"
            )
        if region_def.block_args is not None:
            if len(op.regions[idx].blocks) == 0:
                raise Exception(
                    f"region {region_name} at position {idx} should have at least one block"
                )
            expected_num_args = len(region_def.block_args)
            num_args = len(op.regions[idx].blocks[0].args)
            if num_args != expected_num_args:
                raise Exception(
                    f"region {region_name} at position {idx} should have {expected_num_args} argument, but got {num_args}"
                )
            for arg_idx, arg_type in enumerate(op.regions[idx].blocks[0].args):
                typ = op.regions[idx].blocks[0].args[arg_idx]
                if arg_type != typ:
                    raise Exception(
                        f"argument at position {arg_idx} in region {region_name} at position {idx} should be of type {arg_type}, but {typ}"
                    )

    for attr_name, attr_def in attributes:
        if attr_name not in op.attributes:
            raise Exception(f"attribute {attr_name} expected")
        attr_def.constr.verify(op.attributes[attr_name])


def irdl_op_definition(cls) -> typing.Type[Operation]:
    """Decorator used on classes to define a new operation definition."""
    operands = [(field_name, field)
                for field_name, field in cls.__dict__.items()
                if isinstance(field, OperandDef)]
    results = [(field_name, field)
               for field_name, field in cls.__dict__.items()
               if isinstance(field, ResultDef)]
    regions = [(field_name, field)
               for field_name, field in cls.__dict__.items()
               if isinstance(field, RegionDef)]
    attributes = [(field_name, field)
                  for field_name, field in cls.__dict__.items()
                  if isinstance(field, AttributeDef)]
    options = cls.__dict__.get("irdl_options", [])
    new_attrs = dict()

    # Add operand access fields
    previous_variadics = 0
    for operand_idx, (operand_name, operand_def) in enumerate(operands):
        new_attrs[operand_name] = property(
            lambda self, idx=operand_idx, previous_vars=previous_variadics:
            get_operand_or_result(self, idx, previous_vars, True))
        if isinstance(operand_def, VarOperandDef):
            previous_variadics += 1
    if previous_variadics > 1 and AttrSizedOperandSegments() not in options:
        raise Exception(
            "Operation defines more than two variadic operands, "
            "but do not define the AttrSizedOperandSegments option")

    # Add result access fields
    previous_variadics = 0
    for result_idx, (result_name, result_def) in enumerate(results):
        new_attrs[result_name] = property(
            lambda self, idx=result_idx, previous_vars=previous_variadics:
            get_operand_or_result(self, idx, previous_vars, False))
        if isinstance(result_def, VarResultDef):
            previous_variadics += 1
    if previous_variadics > 1 and AttrSizedResultSegments() not in options:
        raise Exception("Operation defines more than two variadic results, "
                        "but do not define the AttrSizedResultSegments option")

    for region_idx, (region_name, _) in enumerate(regions):
        new_attrs[region_name] = property(
            (lambda idx: lambda self: self.regions[idx])(region_idx))

    for attribute_name, _ in attributes:
        new_attrs[attribute_name] = property(
            (lambda name: lambda self: self.attributes[name])(attribute_name))

    new_attrs["irdl_operand_defs"] = operands
    new_attrs["irdl_result_defs"] = results
    new_attrs["irdl_region_defs"] = regions
    new_attrs["irdl_attribute_defs"] = attributes
    new_attrs["irdl_options"] = options

    new_attrs["verify_"] = lambda op: irdl_op_verify(op, operands, results,
                                                     regions, attributes)
    if "verify_" in cls.__dict__:
        custom_verifier = cls.__dict__["verify_"]

        def new_verifier(verifier, op):
            verifier(op)
            custom_verifier(op)

        new_attrs["verify_"] = (
            lambda verifier: lambda op: new_verifier(verifier, op))(
                new_attrs["verify_"])

    return type(cls.__name__, (Operation, ), {**cls.__dict__, **new_attrs})


@dataclass
class ParameterDef:
    """An IRDL definition of an attribute parameter."""
    constr: AttrConstraint

    def __init__(self, typ: Union[Attribute, typing.Type[Attribute],
                                  AttrConstraint]):
        self.constr = attr_constr_coercion(typ)


def irdl_attr_verify(attr: ParametrizedAttribute,
                     parameters: List[ParameterDef]):
    """Given an IRDL definition, verify that an attribute satisfies its invariants."""

    if len(attr.parameters) != len(parameters):
        raise Exception(
            f"{len(parameters)} parameters expected, got {len(attr.parameters)}"
        )
    for idx, param_def in enumerate(parameters):
        param_def.constr.verify(attr.parameters[idx])


def irdl_attr_definition(cls) -> typing.Type[Attribute]:
    """Decorator used on classes to define a new attribute definition."""

    parameters = []
    new_attrs = dict()
    for field_name in cls.__dict__:
        field_ = cls.__dict__[field_name]
        if isinstance(field_, ParameterDef):
            new_attrs[field_name] = property(
                (lambda idx: lambda self: self.parameters[idx])(
                    len(parameters)))
            parameters.append(field_)

    new_attrs["verify"] = lambda typ: irdl_attr_verify(typ, parameters)

    if "verify" in cls.__dict__:
        custom_verifier = cls.__dict__["verify"]

        def new_verifier(verifier, op):
            verifier(op)
            custom_verifier(op)

        new_attrs["verify"] = (
            lambda verifier: lambda op: new_verifier(verifier, op))(
                new_attrs["verify"])

    return type(cls.__name__, (ParametrizedAttribute, ), {
        **cls.__dict__,
        **new_attrs
    })
